/**
  ******************************************************************************
  * @file    usbd_audio.c
  * @author  MK
  * @version V0.2
  * @date    08/01/2021
  * @brief   This file provides common functions for USB Audio Class 1&2.
  *
  ******************************************************************************
  */ 
#include "main.h"
#include "usbd_audio.h"
#include "uac2_descriptor.h"
#include <stdbool.h>
#include "usb_util.h"
#include "adc.h"

/* USB Standard Device Descriptor */
__ALIGN_BEGIN static uint8_t USBD_AUDIO_DeviceQualifierDesc[USB_LEN_DEV_QUALIFIER_DESC] __ALIGN_END=
{
  USB_LEN_DEV_QUALIFIER_DESC,      /* bLength */
  USB_DESC_TYPE_DEVICE_QUALIFIER,  /* bDescriptorType */
  0x00,			                   /* bcdUSB */
  0x02,
  0xef,                            /* bDeviceClass    0x00 */
  0x02,                            /* bDeviceSubClass 0x00 */
  0x01,                            /* bDeviceProtocol 0x00 */
  0x40,                            /* bMaxPacketSize */
  USBD_MAX_NUM_CONFIGURATION,      /* bNumConfigurations */
  0x00,                            /* bReserved */
};

uint8_t *uac_cfg_desc;
uint16_t uac_cfg_desc_size;

// Samples per IN packet
uint32_t in_samples __attribute__ ((aligned(4))) = IN_SAMPLES_PER_PACKET(USBD_AUDIO_FREQ_DEFAULT);
uint32_t in_samples_min __attribute__ ((aligned(4)));
uint32_t in_samples_max __attribute__ ((aligned(4)));

uint32_t rx_buffer[AUDIO_IN_SAMPLE_BUF_SIZE * 2] __attribute__ ((aligned(4)));
uint8_t inEmptyBuffer[AUDIO_IN_MAX_PACKET_SIZE] __attribute__ ((aligned(4)));
static uint8_t usb_packet[AUDIO_IN_MAX_PACKET_SIZE] __attribute__ ((aligned(4)));

extern USBD_HandleTypeDef hUsbDevice;

/**
  * @brief  USBD_AUDIO_Init
  *         Initialize the AUDIO interface
  * @param  pdev: device instance
  * @param  cfgidx: Configuration index
  * @retval status
  */
uint8_t  USBD_AUDIO_Init (USBD_HandleTypeDef *pdev, uint8_t cfgidx)
{
	UNUSED(cfgidx);
	USBD_AUDIO_DeviceTypeDef *adev;

#ifdef USE_USB_FS
	USB_OTG_GlobalTypeDef* USBx = USB_OTG_FS;
#else
	USB_OTG_GlobalTypeDef* USBx = USB_OTG_HS;
#endif
	if (_debug_output) {
		printf("Init: %02x\r\n", cfgidx);
		printf("GUSBCFG: %08lx\r\n", USBx->GUSBCFG);
	}
	uac_cfg_desc = UAC2_CfgDesc;
	uac_cfg_desc_size = UAC2_CONFIG_DESC_SIZE;

	/* Allocate Audio structure */
	adev = USBD_malloc(sizeof(USBD_AUDIO_DeviceTypeDef));
	if (adev == NULL) {
		pdev->pClassData = NULL;
	    return (uint8_t)USBD_EMEM;
	}
	adev->in.buffer = (uint32_t *)USBD_malloc(AUDIO_IN_TOTAL_BUF_SIZE * 4);
	if (adev->in.buffer == NULL) {
		USBD_free(adev->out.buffer);
		USBD_free(pdev->pClassData);
		pdev->pClassData = NULL;
	    return (uint8_t)USBD_EMEM;
	}
	/* Initialize audio interfaces */
	adev->in.wr_ptr = 0;
	adev->in.rd_ptr = 0;
	adev->in.fs = USBD_AUDIO_FREQ_DEFAULT;
	adev->in.bytes_per_sample = 2U;
	adev->in.alt_setting = 0xff;
	adev->in.state = IDLE;
	adev->in.parity = PARITY_UNDEF;

	pdev->pClassData = (void *)adev;
	if (pdev->dev_speed == USBD_SPEED_HIGH) {
		pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].bInterval = AUDIO_HS_BINTERVAL;
	}
	else {   /* LOW and FULL-speed endpoints */
	    pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].bInterval = AUDIO_FS_BINTERVAL;
	}

	/* Open EP IN */
	(void)USBD_LL_OpenEP(pdev, AUDIO_IN_EP, USBD_EP_TYPE_ISOC, AUDIO_IN_MAX_PACKET_SIZE);
	pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].is_used = 1U;
	/* ## Flush EP */
	USBD_LL_FlushEP(pdev, AUDIO_IN_EP);

    return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_DeInit
  *         DeInitialize the AUDIO layer
  * @param  pdev: device instance
  * @param  cfgidx: Configuration index
  * @retval status
  */
uint8_t  USBD_AUDIO_DeInit (USBD_HandleTypeDef *pdev, uint8_t cfgidx)
{
	/* Close EP IN */
	(void)USBD_LL_CloseEP(pdev, AUDIO_IN_EP);
	pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].is_used = 0U;
	pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].bInterval = 0U;

	/* DeInit  physical Interface components */
	if(pdev->pClassData != NULL) {
		((USBD_AUDIO_ItfTypeDef *)pdev->pUserData)->DeInit(0);
		USBD_free(((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in.buffer);
		USBD_free(pdev->pClassData);
		pdev->pClassData = NULL;
	}
	return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_GetCfgDesc
  *         return configuration descriptor
  * @param  speed : current device speed
  * @param  length : pointer data length
  * @retval pointer to descriptor buffer
  */
uint8_t  *USBD_AUDIO_GetCfgDesc (uint16_t *length)
{
	*length = (uint16_t)sizeof (UAC2_CfgDesc);
    if (_debug_output) {
    	printf("GetCfgDesc: %d\r\n", *length);
    }
	return UAC2_CfgDesc;
}

/**
* @brief  DeviceQualifierDescriptor
*         return Device Qualifier descriptor
* @param  length : pointer data length
* @retval pointer to descriptor buffer
*/
uint8_t *USBD_AUDIO_GetDeviceQualifierDesc(uint16_t *length)
{
	*length = (uint16_t)sizeof(USBD_AUDIO_DeviceQualifierDesc);
    if (_debug_output) {
	    printf("GetDevDesc: %d\r\n", *length);
    }
	return USBD_AUDIO_DeviceQualifierDesc;
}

/**
* @brief  USBD_AUDIO_RegisterInterface
* @param  fops: Audio interface callback
* @retval status
*/
uint8_t USBD_AUDIO_RegisterInterface(USBD_HandleTypeDef *pdev,
                                     USBD_AUDIO_ItfTypeDef *fops)
{
	if (fops == NULL) {
		return (uint8_t)USBD_FAIL;
	}
	pdev->pUserData = fops;
	return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_SOF
  *         handle SOF event
  * @param  pdev: device instance
  * @retval status
  * SOF handler sends data to the synchronization endpoint
  */
uint8_t USBD_AUDIO_SOF (USBD_HandleTypeDef *pdev)
{
	return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_IsoINIncomplete
  *         handle data ISO IN Incomplete event
  * @param  pdev: device instance
  * @param  epnum: endpoint index
  * @retval status
  */
uint8_t  USBD_AUDIO_IsoINIncomplete (USBD_HandleTypeDef *pdev, uint8_t epnum)
{
	// Check all IN-endpoints
	for (int i = 1; i <= USBD_AUDIO_MAX_IN_EP; i++) {
	    if (USB_DIEPCTL(i) & USB_OTG_DIEPCTL_EPENA) {
	    	if (epnum == (AUDIO_IN_EP & EP_ADDR_MSK)) {
	    		// Flip parity if not flipped already
	    	   	USBD_AUDIO_HandleTypeDef *haudio;
    	    	haudio = &((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in;
	    		if (haudio->parity != PARITY_FLIP) {
	    	    	if (_debug_output) {
	    	    		printf("diepctl2: %d %08lx %ld\r\n", i, USB_DIEPCTL(epnum), HAL_GetTick());
	    	    	}
	    			FLIP_PARITY(epnum);
	    			USB_CLEAR_INCOMPLETE_IN_EP(epnum);
	    			USBD_LL_FlushEP(pdev, AUDIO_IN_EP);
	    			USBD_LL_Transmit(pdev, AUDIO_IN_EP, inEmptyBuffer, AUDIO_IN_PACKET_SIZE(haudio->fs, haudio->bytes_per_sample));
	    			haudio->parity = PARITY_FLIP;
	    		}
	    	}
	    }
	}
    return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_IsoOutIncomplete
  *         handle data ISO OUT Incomplete event
  * @param  pdev: device instance
  * @param  epnum: endpoint index
  * @retval status
  */
uint8_t USBD_AUDIO_IsoOutIncomplete(USBD_HandleTypeDef *pdev, uint8_t epnum)
{
	if (epnum != 0) {
		if (_debug_output) {
			printf("IsoOut: %d %ld\r\n", epnum, HAL_GetTick());
		}
	}
	UNUSED(pdev);
	UNUSED(epnum);
	return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_DataIn
  *         handle data IN Stage
  * @param  pdev: device instance
  * @param  epnum: endpoint index
  * @retval status
  */
uint8_t  USBD_AUDIO_DataIn (USBD_HandleTypeDef *pdev, uint8_t epnum)
{
	if (epnum == (AUDIO_IN_EP & EP_ADDR_MSK)) {
		static uint16_t packet_counter = 0;
    	USBD_AUDIO_HandleTypeDef *haudio;
    	haudio = &((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in;
    	uint16_t samples = AUDIO_IN_GetPacketSizeInSamples(haudio->fs, packet_counter);
    	packet_counter++;
    	if (haudio->state == IDLE) {
    		AUDIO_IN_StartRecord((uint8_t *)rx_buffer, AUDIO_IN_SAMPLE_BUF_SIZE * 2);
    		haudio->state = STARTED;
    		// Send empty buffer to host
        	USBD_LL_Transmit(pdev, AUDIO_IN_EP, inEmptyBuffer, samples * 2 * haudio->bytes_per_sample);
    	}
    	else if (haudio->state == STOPPED) {
     		// Keep sending empty buffer to host
        	USBD_LL_Transmit(pdev, AUDIO_IN_EP, inEmptyBuffer, samples * 2 * haudio->bytes_per_sample);
    	}
    	else if (haudio->state == STREAMING) {
    		volatile int16_t gap = haudio->wr_ptr - haudio->rd_ptr;
    		if (gap < 0) {
    			gap += AUDIO_IN_TOTAL_BUF_SIZE;
    		}
       		if (gap >= samples * 2) {
    			if (gap > IN_GAP_HIGH) {
    				if (samples < in_samples_max) {
    					samples++;
    				}
    			} else if (gap < IN_GAP_LOW) {
    				if (samples > in_samples_min) {
    					samples--;
    				}
     			}
    			uint16_t packet_size = samples * 2;
    			uint16_t remaining = AUDIO_IN_TOTAL_BUF_SIZE - haudio->rd_ptr;
     			if (remaining < packet_size) {
    				AUDIO_IN_CopySamplesToUsbPacket(usb_packet, &haudio->buffer[haudio->rd_ptr], remaining, haudio->bytes_per_sample);
    				AUDIO_IN_CopySamplesToUsbPacket(&usb_packet[remaining*haudio->bytes_per_sample], haudio->buffer, packet_size - remaining, haudio->bytes_per_sample);
    			}
    			else {
    				AUDIO_IN_CopySamplesToUsbPacket(usb_packet, &haudio->buffer[haudio->rd_ptr], packet_size, haudio->bytes_per_sample);
    			}
				USBD_LL_Transmit(pdev, AUDIO_IN_EP, usb_packet, packet_size * haudio->bytes_per_sample);
				haudio->rd_ptr = (haudio->rd_ptr + packet_size) % AUDIO_IN_TOTAL_BUF_SIZE;
    		}
    		else {
    			// Not enough data. Send remaining samples to host and stop recording
    			if (_debug_output) {
    				printf("IN eod: %d %d %d %d\r\n", gap, samples, haudio->wr_ptr, haudio->rd_ptr);
    			}
    			USBD_LL_Transmit(pdev, AUDIO_IN_EP, (uint8_t*) (&haudio->buffer[haudio->rd_ptr]), gap * 2 * haudio->bytes_per_sample);
    			haudio->state = STOPPED;
    		}
    	}
    	else {
    		// Send empty buffer
    		USBD_LL_Transmit(pdev, AUDIO_IN_EP, inEmptyBuffer, samples * 2 * haudio->bytes_per_sample);
    	}
    }
    return (uint8_t)USBD_OK;
}

/**
  * @brief  USBD_AUDIO_DataOut
  *         handle data OUT Stage
  * @param  pdev: device instance
  * @param  epnum: endpoint index
  * @retval status
  */
uint8_t USBD_AUDIO_DataOut(USBD_HandleTypeDef *pdev, uint8_t epnum)
{
  	return (uint8_t)USBD_OK;
}

/**
  * @brief  AUDIO_IN_DMA_Sync
  * @param  pdev: device instance;
  * @param  offset: input buffer offset
  * @retval status
  */
void AUDIO_IN_DMA_Sync(USBD_HandleTypeDef *pdev, AUDIO_OffsetTypeDef offset)
{
	USBD_AUDIO_HandleTypeDef *haudio;
	haudio = &((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in;

	// Fill ring buffer with data received from USB
	if (offset == AUDIO_OFFSET_HALF) {
		AUDIO_IN_CopyToRingBuffer(haudio, rx_buffer, AUDIO_IN_SAMPLE_BUF_SIZE);
	}
	else if (offset == AUDIO_OFFSET_FULL) {
		AUDIO_IN_CopyToRingBuffer(haudio, &rx_buffer[AUDIO_IN_SAMPLE_BUF_SIZE], AUDIO_IN_SAMPLE_BUF_SIZE);
	}
	// Enable sending to host when buffer is half full
	if (haudio->state == STARTED) {
		if (haudio->wr_ptr - haudio->rd_ptr > IN_GAP_NOM) {
			haudio->state = STREAMING;
	  		if (_debug_output) {
	  			printf("Record: %d %d\r\n", haudio->wr_ptr, haudio->rd_ptr);
	  		}
		}
	}
}

/**
  * @brief  AUDIO_OUT_DMA_Sync
  * @param  pdev: device instance;
  * @param  offset: output buffer offset
  * @retval status
  */
void AUDIO_OUT_DMA_Sync(USBD_HandleTypeDef *pdev, AUDIO_OffsetTypeDef offset)
{
}

/**
 * @brief  Return USB packet size in samples
 * @param  fs: sample rate
 * @param  counter: packet counter
 */
uint16_t AUDIO_IN_GetPacketSizeInSamples(uint32_t fs, uint16_t counter)
{
	uint8_t samples = in_samples;

	switch (fs) {
		case 44100L:
			if (counter == 80) {
				counter = 0;
			}
			else if (counter % 2 == 0 || counter >= 79) {
				samples = in_samples_max;
			}
			break;
		case 88200L:
			if (counter == 40) {
				counter = 0;
			}
			else if (counter >= 39) {
				samples = in_samples_max;
			}
			break;
		case 176400L:
			if (counter == 20) {
				counter = 0;
			}
			else if (counter >= 19) {
				samples = in_samples_max;
			}
			break;
		case 352800L:
			if (counter == 10) {
				counter = 0;
			}
			else if (counter >= 9) {
				samples = in_samples_max;
			}
			break;
		case 705600L:
			if (counter == 5) {
				counter = 0;
			}
			else if (counter >= 4) {
				samples = in_samples_max;
			}
			break;
		case 39040L:
			if (counter == 25) {
				counter = 0;
			}
			else if (counter%8 != 0) {
				samples = in_samples_max;
			}
			break;
	}
	return samples;
}

/**
 * @brief  Set audio input sample rate.
 * @param  sample_rate: new sample rate
 */
void AUDIO_IN_SetSampleRate(USBD_HandleTypeDef* pdev, uint32_t sample_rate)
{
	uint8_t index = 0;
	USBD_AUDIO_HandleTypeDef* haudio;

	haudio = &((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in;
#ifdef USE_USB_FS
	const uint32_t IN_SAMPLES_PER_PACKET_TABLE[][4]=
	{
		 {44100L,  44,  44,  45 }, //44.1kHz
		 {48000L,  48,  47,  49 }, //48kHz
		 {88200L,  88,  88,  89 }, //88.2kHz
		 {96000L,  96,  95,  97 }, //96kHz
		 {176400L, 176, 176, 177 }, //176.4kHz
		 {192000L, 192, 191, 193 }, //192kHz
	};
#else	// HS
	const uint32_t IN_SAMPLES_PER_PACKET_TABLE[][4]=
	{
		 {44100L,  5,  5,  6 }, //44.1kHz
		 {48000L,  6,  5,  7 }, //48kHz
		 {88200L,  11, 11, 12 }, //88.2kHz
		 {96000L,  12, 11, 13 }, //96kHz
		 {176400L, 22, 22, 23 }, //176.4kHz
		 {192000L, 24, 23, 25 }, //192kHz
		 {352800L, 44, 44, 45 }, //352.8kHz
		 {384000L, 48, 47, 49 }, //384kHz
		 {705600L, 88, 88, 89 }, //705.6kHz
		 {768000L, 96, 95, 97 }, //768kHz
		 {39040L,  4,  4,  5 }, //39.1kHz
	};
#endif
	switch (sample_rate) {
    	case 44100L:
    		index = 0;
    		break;
    	case 48000L:
    		index = 1;
    		break;
    	case 88200L:
    		index = 2;
    		break;
    	case 96000L:
    		index = 3;
    		break;
    	case 176400L:
    		index = 4;
			break;
    	case 192000L:
    		index = 5;
     		break;
    	case 352800L:
    		index = 6;
     		break;
    	case 384000L:
    		index = 7;
     		break;
    	case 705600L:
    		index = 8;
     		break;
    	case 768000L:
    		index = 9;
     		break;
    	case 39040L:
    		index = 10;
    		break;
    	default:
    		break;
	}
	ADC_SetFS(sample_rate);
	in_samples = IN_SAMPLES_PER_PACKET_TABLE[index][1];
	in_samples_min = IN_SAMPLES_PER_PACKET_TABLE[index][2];
	in_samples_max = IN_SAMPLES_PER_PACKET_TABLE[index][3];
	haudio->fs = sample_rate;
	if (_debug_output) {
		printf("Set in sample rate: %ld %d %ld %ld %ld\r\n", sample_rate, index, in_samples, in_samples_min, in_samples_max);
	}
}

/**
 * @brief  Stop recording and reset buffer pointers
 * @param  pdev: instance
 */
void AUDIO_IN_StopAndReset(USBD_HandleTypeDef* pdev)
{
	USBD_AUDIO_HandleTypeDef* haudio;

	AUDIO_IN_StopRecord();

	haudio = &((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in;
	haudio->state = IDLE;
	haudio->rd_ptr = 0U;
	haudio->wr_ptr = 0U;

	USBD_LL_FlushEP(pdev, AUDIO_IN_EP);
	(void)USBD_LL_CloseEP(pdev, AUDIO_IN_EP);
	pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].is_used = 0U;
}

/**
 * @brief  Restart recording with new parameters
 * @param  pdev: instance
 */
void AUDIO_IN_Restart(USBD_HandleTypeDef* pdev)
{
	USBD_AUDIO_HandleTypeDef* haudio;
	haudio = &((USBD_AUDIO_DeviceTypeDef*)pdev->pClassData)->in;
	(void)USBD_LL_OpenEP(pdev, AUDIO_IN_EP, USBD_EP_TYPE_ISOC, AUDIO_IN_MAX_PACKET_SIZE);
	USBD_LL_FlushEP(pdev, AUDIO_IN_EP);
	pdev->ep_in[AUDIO_IN_EP & EP_ADDR_MSK].is_used = 1U;
   	USBD_LL_Transmit(pdev, AUDIO_IN_EP, inEmptyBuffer, AUDIO_IN_PACKET_SIZE(haudio->fs, haudio->bytes_per_sample));
	haudio->state = IDLE;
	haudio->parity = PARITY_UNDEF;
}

/**
  * @brief  AUDIO_IN_CopySamplesToUsbPacket
  * @param  dst: USB packet
  * @param  src: source buffer
  * @param  samples: size in samples
  * @param  dst_bytes_per_sample: bytes per sample in USB packet
  */
void AUDIO_IN_CopySamplesToUsbPacket(uint8_t *dst, uint32_t *src, uint16_t samples, uint8_t dst_bytes_per_sample)
{
	if (dst_bytes_per_sample == 2) {
		uint16_t *dstptr = (uint16_t *)dst;
		for (int i = 0; i < samples; i++) {
			dstptr[i] = (uint16_t)(src[i] >> 16);
		}
	}
	else if (dst_bytes_per_sample == 3) {
		uint8_t *dptr = (uint8_t *)dst;
		uint8_t *sptr = (uint8_t *)src;
		for (int i = 0; i < samples; i++) {
			sptr++;
			for (int j = 0; j < 3; j++) {
				*dptr++ = *sptr++;
			}
		}
	}
	else if (dst_bytes_per_sample == 4) {
		uint32_t *dstptr = (uint32_t *)dst;
		for (int i = 0; i < samples; i++) {
			dstptr[i] = src[i];
		}
	}
}

/**
  * @brief  AUDIO_IN_CopySamples
  * @param  dst: destination buffer (contains samples in host datasize)
  * @param  src: source buffer (contains samples in SAI datasize)
  * @param  dst_bytes_per_sample: destination datasize
  */
void AUDIO_IN_CopySamples(uint32_t *dst, uint32_t *src, uint16_t samples)
{
	DMA2->LIFCR = (uint32_t)0x003B0000;      // clear DMA low IRQ flags
	DMA2->HIFCR = (uint32_t)0x003B0000;      // clear DMA high IRQ flags
	DMA2_Stream2->CR |= (0b10 << DMA_SxCR_DIR_Pos) // Memory-to-memory Mode
	            | (0b10 << DMA_SxCR_PSIZE_Pos) // Source: 32-bit
	            | (0b10 << DMA_SxCR_MSIZE_Pos) // Destination: 32-bit
	            | DMA_SxCR_PINC // Increment source pointer
	            | DMA_SxCR_MINC; // Increment destination pointer
	DMA2_Stream2->PAR = (uint32_t) src;
	DMA2_Stream2->M0AR = (uint32_t) dst;
	DMA2_Stream2->NDTR = (uint16_t)samples;
	DMA2_Stream2->CR |= DMA_SxCR_EN; // Start DMA transfer
}

/**
  * @brief  AUDIO_IN_CopyToRingBuffer
  * @param  haudio: audio device handle
  * @param  src: source buffer
  * @param  samples: size in samples
  */
void AUDIO_IN_CopyToRingBuffer(USBD_AUDIO_HandleTypeDef *haudio, uint32_t *src, uint16_t samples)
{
  	uint16_t remaining = AUDIO_IN_TOTAL_BUF_SIZE - haudio->wr_ptr;
  	if (remaining < samples) {
  		if (remaining > 0) {
  			AUDIO_IN_CopySamples(&haudio->buffer[haudio->wr_ptr], src, remaining);
  			haudio->wr_ptr = 0;
  			samples -= remaining;
  		}
  		if (samples > 0) {
  			AUDIO_IN_CopySamples(&haudio->buffer[haudio->wr_ptr], &src[remaining], samples);
  			haudio->wr_ptr += samples;
  		}
  	}
  	else {
  		if (samples > 0) {
  			AUDIO_IN_CopySamples(&haudio->buffer[haudio->wr_ptr], src, samples);
  			haudio->wr_ptr += samples;
  		}
  	}
  	// reset write pointer if end of buffer reached
  	if (haudio->wr_ptr >= AUDIO_IN_TOTAL_BUF_SIZE) {
  		haudio->wr_ptr = 0;
  	}
}

/**
  * @brief  AUDIO_IN_StartRecord
  */
void AUDIO_IN_StartRecord(uint8_t *buf, uint16_t size)
{
	HAL_SAI_Receive_DMA(&hsai_in, buf, size);
}

/**
  * @brief  AUDIO_IN_StopRecord
  */
void AUDIO_IN_StopRecord(void)
{
	if (_debug_output) {
		printf("Stop record\r\n");
	}
	// Stop DMA transfer
	HAL_SAI_DMAStop(&hsai_in);
	// Flush the DMA fifo
	SET_BIT(hsai_in.Instance->CR2, SAI_xCR2_FFLUSH);
}


/**
  * @brief  Manages the RX DMA full Transfer complete event.
  * @param  None
  * @retval None
  */
void HAL_SAI_RxCpltCallback(SAI_HandleTypeDef *hsai)
{
	if (hsai->Instance == hsai_in.Instance) {
		USBD_AUDIO_HandleTypeDef *haudio;
		haudio = &((USBD_AUDIO_DeviceTypeDef*)hUsbDevice.pClassData)->in;
		switch(haudio->state) {
			case STARTED:
			case STREAMING:
				AUDIO_IN_DMA_Sync(&hUsbDevice, AUDIO_OFFSET_FULL);
				break;
			case STOPPED:
				haudio->state = IDLE;
				break;
			case IDLE:
				break;
		}
	}
}

/**
  * @brief  Manages the RX DMA Half Transfer complete event.
  * @param  None
  * @retval None
  */
void HAL_SAI_RxHalfCpltCallback(SAI_HandleTypeDef *hsai)
{
	if (hsai->Instance == hsai_in.Instance) {
		USBD_AUDIO_HandleTypeDef *haudio;
		haudio = &((USBD_AUDIO_DeviceTypeDef*)hUsbDevice.pClassData)->in;
		switch(haudio->state) {
			case STARTED:
			case STREAMING:
				AUDIO_IN_DMA_Sync(&hUsbDevice, AUDIO_OFFSET_HALF);
				break;
			case STOPPED:
				// do nothing
				break;
			case IDLE:
				// do nothing
				break;
		}
	}
}

/**
  * @brief SAI error callback.
  * @param  hsai pointer to a SAI_HandleTypeDef structure that contains
  *                the configuration information for SAI module.
  * @retval None
  */
void HAL_SAI_ErrorCallback(SAI_HandleTypeDef *hsai)
{
	if (_debug_output) {
		printf("SAI error: %08lx\r\n", hsai->ErrorCode);
	}
}
